$NOLIST
;
;********************************************************************************************************
; Ziehen der Quadratwurzel (32 Bit -> 16 Bit)
; J. Rathlev, Dez. 2019
; Diese Datei kann mit der Include-Anweisung in andere Programme integriert werden
;********************************************************************************************************

        DSEG

DRADI:   DS      4       ;  Little-Endian-Format 
DSQTMP:  DS      7
DROOT:   DS      2 

;********************************************************************************************************

        CSEG

;********************************************************************************************************
; RADI nach links schieben, herausgeschobene Bits in SQTMP, von rechts wird mit 0 aufgefllt
; benutzte Register: A,R7
;********************************************************************************************************

DSH162: MOV     R7,#2           ; 2 Bit
DSH16L: CLR     C
        MOV     A,DRADI
        RLC     A
        MOV     DRADI,A
        MOV     A,DRADI+1
        RLC     A
        MOV     DRADI+1,A
        MOV     A,DRADI+2
        RLC     A
        MOV     DRADI+2,A
        MOV     A,DRADI+3
        RLC     A
        MOV     DRADI+3,A
        MOV     A,DSQTMP
        RLC     A
        MOV     DSQTMP,A
        MOV     A,DSQTMP+1
        RLC     A
        MOV     DSQTMP+1,A
        MOV     A,DSQTMP+2
        RLC     A
        MOV     DSQTMP+2,A
        MOV     A,DSQTMP+3
        RLC     A
        MOV     DSQTMP+3,A
        DJNZ    R7,DSH16L
        RET
        

;********************************************************************************************************
; Ziehen der Quadratwurzel (ca. 810 s bei 22,1184 MHz)
; Prinzip: (b[7]+b[6]+b[5]+..+b[0])^2 = b[7]^2 + (2*b[7]+b[6])*b[6] + (2*(b[7]+b[6])+b[5])*b[5] + ..
; (b[7] = 2^7 = 128 oder 0, b[6] = 2^6 = 64 oder 0, .. , b[0] = 2^0 = 1 oder 0)
; Radikant in DRADI (32 Bit) - Little-Endian-Format
; Ergebnis in DROOT (16 Bit)
; benutzte Register: A,R7,R6,R0
;********************************************************************************************************

SQRT16: CLR     A
        MOV     R7,#9
        MOV     R0,#DSQTMP
SQ16L:  MOV     @R0,A           ; DSQTMP und DROOT auf 0 setzen
        INC     R0
        DJNZ    R7,SQ16L

        MOV     R6,#15          ; Zhler
        CALL    DSH162          ; 2 Bit links
        MOV     A,DSQTMP
        JZ      SQ161           ; 1. Ergebnisbit = 0
        INC     DROOT            ; 1. Ergebnisbit = 1
        SUBB    A,DROOT
        MOV     DSQTMP,A
SQ161:  CALL    DSH162          ; 2 Bit links
        MOV     A,DROOT
        CLR     C
        RLC     A               ; 1 Bit links
        MOV     DROOT,A
        MOV     A,DROOT+1
        RLC     A               ; 1 Bit links
        MOV     DROOT+1,A
        MOV     A,DROOT
        SETB    C               ; nchstes Ergebnisbit probeweise setzen
        RLC     A               ; (2*a+b)*b
        MOV     DSQTMP+4,A       ; zwischenspeichern
        MOV     A,DROOT+1
        RLC     A
        MOV     DSQTMP+5,A       ; zwischenspeichern
        CLR     A
        RLC     A               
        MOV     DSQTMP+6,A       ; zwischenspeichern
        
        MOV     A,DSQTMP
        SUBB    A,DSQTMP+4
        MOV     R7,A             ; zwischenspeichern
        MOV     A,DSQTMP+1
        SUBB    A,DSQTMP+5
        MOV     R0,A
        MOV     A,DSQTMP+2
        SUBB    A,DSQTMP+6
        JC      SQ162            ; Bit nicht setzen
        
        MOV     DSQTMP+2,A       ; Subtraktionsergebnis speichern
        MOV     DSQTMP+1,R0   
        MOV     DSQTMP,R7
        INC     DROOT            ; Bit setzen
        MOV     A,DROOT
        JNZ     SQ162
        INC     DROOT+1    
SQ162:  DJNZ    R6,SQ161
        RET
                    
; Bytes umspeichern von Programmspeicher (DPTR) nach Datenspeicher (R0)
; MOVCD -  Anzahl der Bytes in R7
; MOVCD2 - 2 Bytes  (11,5 s bei 22 MHz)
; MOVCD4 - 4 Bytes  (21,5 s bei 22 MHz)

MOVCD4: MOV     R7,#4
        SJMP    MOVCD
MOVCD2: MOV     R7,#2            
MOVCD:  CLR     A
        MOVC    A,@A+DPTR
        INC     DPTR
        MOV     @R0,A
        INC     R0
        DJNZ    R7,MOVCD
        RET

