$NOLIST

; Include Datei

;********************************************************************************************************
; Abfrage einer Zehnertastatur mit zwei Sondertasten (7-pol. Anschluss)
; ber einen I2C-Portexpander
; J. Rathlev, Nov. 2006

; Erforderlich:  Include-Datei "I2C.a51"
; Interrupt: Int0 bei Tastenbettigung

; Anschlsse:
;   Tast.-Pin  E-Port     
;      1         6      
;      2         5
;      3         4
;      4         3           
;      5         2
;      6         1
;      7         0           
;     
; Abfrage:            
; Wert    verb. Tastatur-Pins
;   1        1-4       
;   2        2-4       
;   3        3-4       
;   4        1-5       
;   5        2-5       
;   6        3-5       
;   7        1-6       
;   8        2-6       
;   9        3-6       
;   *        1-7       
;   0        2-7       
;   #        3-7

; E-Port 0,1,2,3 (Ausgang) ==> 4,5,6   (Eingang) : Reihen
;        4,5,6   (Ausgang) ==> 0,1,2,3 (Eingang) : Spalten 

;********************************************************************************************************
        DSEG

; RAM-Daten
; =========

TaSAdr: DS      1               ; Subadresse des Portexpanders 
TaOut:  DS      1
TaDat:  DS      2
                                                                                     
        BSEG

TaFlag:  DBIT    1               ; Flag fr erfolgte Tastatureingabe


;********************************************************************************************************
        CSEG

; Tastatur-Interrupt
; ==================

SVPC    SET     $

        ORG     CODEMEM+ExtI0

; Interrupt-Routine
        SETB    TaFlag
        RETI

        ORG     SVPC
                
;********************************************************************************************************
; Initialisieren
; Im Hauptprogramm muss noch folgen:
;         SETB    EA              ; Enable Global-Interrupt
; Subadresse des Portexpanders in A

InitKey:
        MOV     TaSAdr,A
        MOV     TaOut,#0FH      ; Bits 4,5,6 = 0 - 0,1,2,3 = Eingang 
        MOV     R0,#TaOut
        MOV     R1,#1
        MOV     R2,TaSAdr
        CALL    PE0Write
        CLR     TaFlag
        SETB    IT0
        SETB    EX0             ; Int0 von Tastatur
        RET        

;********************************************************************************************************
; Zuordnung der Tasten zu den ASCII-Codes

TTable: DB '147*2580369#'

; Prfen, ob Eingabe erfolgt ist, und Umsetzen der Eingabe in ASCII
; Benutzt:      A,R0,R1,R2,R3,R6,R7,B,DPTR
; Ergebnis:     C = 1 : Eingabe erfolgt
;               A = Eingabewert
;               C = 0 : keine Eingabe

GetKey: JNB     TaFlag,GkRet    ; keine Eingabe erfolgt
        CLR     TaFlag

        MOV     R0,#TaDat
        MOV     R1,#1
        MOV     R2,TaSAdr
        CALL    PE0Read         ; Bits 0,1,2,3 einlesen (Reihe)
        
        MOV     TaOut,#70H      ; Bits 0,1,2,3 = 0 - 4,5,6 = Eingang 
        MOV     R0,#TaOut
        MOV     R1,#1
        MOV     R2,TaSAdr
        CALL    PE0Write
        
        MOV     R0,#TaDat+1
        MOV     R1,#1
        CALL    PE0Read         ; Bits 4,5,6 einlesen (Spalte)
        
        MOV     TaOut,#0FH      ; Bits 4,5,6 = 0 - 0,1,2,3 = Eingang 
        MOV     R0,#TaOut
        MOV     R1,#1
        CALL    PE0Write

        MOV     A,TaDat         ; Reihenbits
        MOV     R3,#4
GkLp1:  RRC     A
        JNC     GkLp4           ; Bit = 0 ?
        DJNZ    R3,GkLp1
        SJMP    GkRet           ; keine Taste gedrckt 
                
GkLp4:  MOV     A,TaDat+1       ; 3 Spaltenbits
        SWAP    A
        MOV     R6,#3     
GkLp2:  RRC     A
        JNC     GkLp3
        DJNZ    R6,GkLp2
        SJMP    GkRet           ; keine Taste gedrckt 

GkLp3:  MOV     A,R6            ; Spalte * 4
        DEC     A
        RL      A
        RL      A
        ADD     A,R3            ; + Reihe
        DEC     A
        MOV     DPTR,#TTable
        MOVC    A,@A+DPTR        
        
GkLp6:  SETB    C
        RET

GkRet:  CLR     C
        RET

$LIST
